#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <curl/curl.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <libconfig.h>

#include "led.h"
#include "logging.h"


const char INFLUXURL_KEY[] = "influxUrl";
const char DEFAULT_INFLUXURL[] = "http://172.16.3.15:8086/write?db=smarthome2&precision=ms";
const char *influxUrl;
const char INFLUXUSER_KEY[] = "influxUser";
const char *influxUser;
const char INFLUXPASS_KEY[] = "influxPass";
const char *influxPass;
const char INFLUXTAG_KEY[] = "influxTag";
const char *influxTag;
const char DEFAULT_LOCATION[] = "Essen_DE";
const char LOCATION_KEY[] = "location";
const char *location;

const uint8_t ONE_SECOND_DIVIDER = 50;


extern uint32_t skipped;


// #define BUFSIZE 131070
#define BUFSIZE 65535
// #define BUFSIZE 1024
// char influxBuffer[BUFSIZE];
// char *bufferNextEntry;

typedef struct influxBuffer {
    uint32_t entries;
    uint32_t totalEntries;
    char *nextEntry;
    char buffer[BUFSIZE];
} tInfluxBuffer;

typedef enum {
    PERIOD_20MS = 0,
    PERIOD_1S,
    PERIOD_END
} ePeriod;

tInfluxBuffer *influxBuffers[PERIOD_END];

#define HOSTNAMESIZE 128
char hostname[HOSTNAMESIZE];

static void influxClearBuffer(tInfluxBuffer *influxBuffer, bool initial) {
    memset(influxBuffer->buffer, 0, BUFSIZE);
    influxBuffer->nextEntry = influxBuffer->buffer;
    influxBuffer->entries = 0;
    if (initial) {
        influxBuffer->totalEntries = 0;
    }

}

void influxInit(config_t *pCfg) {
    if (! config_lookup_string(pCfg, INFLUXURL_KEY, &influxUrl)) {
        influxUrl = DEFAULT_INFLUXURL;
    }
    fprintf(stderr, "CONFIG: influxUrl=%s\n", influxUrl);
    if (! config_lookup_string(pCfg, INFLUXUSER_KEY, &influxUser)) {
        influxUser = NULL;
    }
    fprintf(stderr, "CONFIG: influxUser=%s\n", (influxUser == NULL ? "<null>" : influxUser));
    if (! config_lookup_string(pCfg, INFLUXPASS_KEY, &influxPass)) {
        influxPass = NULL;
    }
    fprintf(stderr, "CONFIG: influxPass=%s\n", (influxPass == NULL ? "<null>" : influxPass));
    if (! config_lookup_string(pCfg, LOCATION_KEY, &location)) {
        location = DEFAULT_LOCATION;
    }
    fprintf(stderr, "CONFIG: location=%s\n", location);
    if (! config_lookup_string(pCfg, INFLUXTAG_KEY, &influxTag)) {
        gethostname(hostname, HOSTNAMESIZE);
        influxTag = hostname;
    }
    fprintf(stderr, "CONFIG: influxTag=%s\n", influxTag);
    influxBuffers[PERIOD_20MS] = (tInfluxBuffer*) malloc(sizeof(tInfluxBuffer));
    influxClearBuffer(influxBuffers[PERIOD_20MS], true);
    influxBuffers[PERIOD_1S] = (tInfluxBuffer*) malloc(sizeof(tInfluxBuffer));
    influxClearBuffer(influxBuffers[PERIOD_1S], true);
}



static void influxSendRequest(tInfluxBuffer *influxBuffer) {
    led(E_RED, false);
    led(E_BLUE, true);
    CURL *curl = curl_easy_init();
    if(curl) {
        curl_easy_setopt(curl, CURLOPT_URL, influxUrl);
        if (influxUser && influxPass) {
            curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_DIGEST);
            curl_easy_setopt(curl, CURLOPT_USERNAME, influxUser);
            curl_easy_setopt(curl, CURLOPT_PASSWORD, influxPass);
        }
        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, influxBuffer->buffer);
        CURLcode res = curl_easy_perform(curl);
        if(res != CURLE_OK) {
            logmsg(LOG_ERR, "curl_easy_perform() failed: %s\n", curl_easy_strerror(res));
            led(E_RED, true);
        }
        curl_easy_cleanup(curl);
    }
    led(E_BLUE, false);
}

void influxAddFrequency(uint32_t period, double fRaw, double fSmoothed, double gradient, 
                        int valid) {
    static uint8_t divider = 0;
    static double summedUpTime = 0.0;
    char tmpBuf[256];
    struct timespec t;

    clock_gettime(CLOCK_REALTIME, &t);
    uint64_t tt = (((uint64_t)t.tv_sec) * 1000) + (((uint64_t)t.tv_nsec) / 1000000);
    int c = sprintf(tmpBuf, "mainsfrequency,host=%s,valid=%d,location=%s "
                            "period=%u,freq=%f,freqSmoothed=%f,gradient=%f,freqRaw=%f "
                            "%llu\n",
                            influxTag, valid, location, 
                            period, fSmoothed, fSmoothed, gradient, fRaw, 
                            tt);

    if ((influxBuffers[PERIOD_20MS]->nextEntry + c + 10) > (influxBuffers[PERIOD_20MS]->buffer + BUFSIZE)) {
        influxSendRequest(influxBuffers[PERIOD_20MS]);
        influxBuffers[PERIOD_20MS]->totalEntries += influxBuffers[PERIOD_20MS]->entries;
        logmsg(LOG_INFO, "%u 20ms-entries sent to database, in total %u, invalid: %u\n", 
            influxBuffers[PERIOD_20MS]->entries, influxBuffers[PERIOD_20MS]->totalEntries, skipped);
        influxClearBuffer(influxBuffers[PERIOD_20MS], false);
    }

    memcpy(influxBuffers[PERIOD_20MS]->nextEntry, tmpBuf, c);
    influxBuffers[PERIOD_20MS]->nextEntry += c;
    influxBuffers[PERIOD_20MS]->entries += 1;

    summedUpTime += fSmoothed;
    divider += 1;

    if (divider == ONE_SECOND_DIVIDER) {
        double freq1S = ((double)ONE_SECOND_DIVIDER) / summedUpTime; 
        int c = sprintf(tmpBuf, "mainsfrequency1S,host=%s,location=%s "
                                "freq=%f "
                                "%llu\n",
                                influxTag, location, 
                                freq1S, 
                                tt);

        //if ((influxBuffers[PERIOD_1S]->nextEntry + c + 10) > (influxBuffers[PERIOD_1S]->buffer + BUFSIZE)) {
        if (influxBuffers[PERIOD_1S]->entries == 10) {
            influxSendRequest(influxBuffers[PERIOD_1S]);
            influxBuffers[PERIOD_1S]->totalEntries += influxBuffers[PERIOD_1S]->entries;
            logmsg(LOG_INFO, "%u 1s-entries sent to database, in total %u\n", 
                influxBuffers[PERIOD_1S]->entries, influxBuffers[PERIOD_1S]->totalEntries);
            influxClearBuffer(influxBuffers[PERIOD_1S], false);
        }

        memcpy(influxBuffers[PERIOD_1S]->nextEntry, tmpBuf, c);
        influxBuffers[PERIOD_1S]->nextEntry += c;
        influxBuffers[PERIOD_1S]->entries += 1;

        divider = 0;
    }
}


